package pep;

import static pep.Main.nGoals;

/**This class defines the operations of Decision nodes.*/
public class Decision extends SubNode{
 /**total amount memory retrieved*/
 public static ThreadLocal<Double> totRetrieved = new ThreadLocal<>();
 /**input for preparedness*/      private double inPrep;
 /**preparedness for activation*/ private double preparedness;

 /**Creates a new Decision node.
  * @param subject Subject number used for static variable array.*/
 public Decision(int subject){
  super(subject);
  threshold = .5;
  strength = 2;
  nConnections = Main.nResponses;
  weight = new double[nConnections];
 }

 /**Updates activation for each cycle and searches responses.*/
 public void activate(){
  retrieve();
  updatePrep();
  input *= preparedness;
  input += addNoise(.3);
  update();
  double output = (activation-threshold);
  if(output>0){
   for(int r=0; r<nConnections;r++)
    Main.runs[subj].responses[r].send(output*weight[r]);
  }
 }

 /**Adds retrieved activation to input with SubNode method.*/
 private void retrieve(){
  double total = totRetrieved.get();
  double multiplier = Executive.decSearch.get();
  retrieve(total,multiplier);
 }

 /**Input to node sensitivity.
  * @param prep Incoming goal activation.*/
 public void prepare(double prep){
  inPrep += range(prep);
 }

 /**Updates node sensitivity to input logistically.*/
 private void updatePrep(){
  inPrep = (1 / (1 + Math.exp(-((inPrep*8)-4))));
  preparedness = preparedness*(1-decay) + inPrep*decay;
  inPrep = .35;
 }

 /**Sends retrieval from Input nodes.
  * @param in Input activation.*/
 public void send(double in){
  retrieved += in;
  double tot = (double)(in/nGoals)+totRetrieved.get();
  totRetrieved.set(tot);
 }

 /**Clears totRetrieved for the next cycle.*/
 public static void clear(){
  totRetrieved.set(0.0);
 }
}