package pep;

import java.text.DecimalFormat;
import static pep.Main.nDistracters; import static pep.Main.nTargets;
import static pep.Main.nCues;        import static pep.Main.nDecisions;
import static pep.Main.nGoals;       import static pep.Main.nResponses;

/**This Node subclass defines the operations of Episode nodes.*/
public class Episode extends Node{
 /**response time*/               private double rt;
 /**starting time weight*/        private double timeWeight;
 /**rate of connection building*/ private static double writeConnect = .01;
 /**decay rate of connections*/   private static double decayConnect = .05;
 /**distracter connection array*/ private double[] distWeight;
 /**target connection array*/     private double[] targWeight;
 /**cue connection array*/        private double[] cueWeight;
 /**decision connection array*/   private double[] decWeight;
 /**goal connection array*/       private double[] goalWeight;
 /**response connection array*/   private double[] respWeight;
 /**peak activation of node*/     private double peakAct;

 /**Creates a new Episode and initializes weight arrays.
  * @param subject Subject number used for static variable array.*/
 public Episode(int subject){
  super(subject);
  threshold = .15;
  distWeight = new double[nDistracters];
  targWeight = new double[nTargets];
  cueWeight = new double[nCues];
  decWeight = new double[nDecisions];
  goalWeight = new double[nGoals];
  respWeight = new double[nResponses];
  timeWeight = .02;
 }

 /**Updates activation for each cycle and searches responses.*/
 public void activate(){
  if(input>.001 || activation>.001){
   update();
   if(activation>peakAct) peakAct = activation;
   if(activation>threshold) respSearch();
  }
 }

 /**Adds input to node (distracters).
  * @param distIn Input from distracter Input node.
  * @param item Item number.*/
 public void memSearchD(double distIn,int item){
  input += (distIn * distWeight[item]);
 }

 /**Adds input to node (targets).
  * @param targIn Input from target Input node.
  * @param item Item number.*/
 public void memSearchT(double targIn,int item){
  input += (targIn * targWeight[item]);
 }

 /**Adds input to node (cues).
  * @param cueIn Input from cue Input node.
  * @param item Item number.*/
 public void memSearchC(double cueIn,int item){
  input += (cueIn * cueWeight[item]);
 }

 /**Adds input to node (goals).
  * @param goalIn Input from Goal node.
  * @param item Item number.*/
 public void memSearchGoal(double goalIn,int item){
  input += (goalIn * goalWeight[item]);
 }

 /**Adds input to node (decisions).
  * @param decIn Input from Decision node.
  * @param item Item number.*/
 public void memSearchDec(double decIn,int item){
  input += (decIn * decWeight[item]);
 }

 /**Adds input to node (responses).
  * @param respIn Input from Response node.
  * @param item Item number.*/
 public void memSearchR(double respIn,int item){
  input += (respIn * respWeight[item]);
 }

 /**Activates connected nodes and sums retrieval totals.*/
 private void respSearch(){
  double outAct = (activation-threshold);
  for(int c=0; c<nCues; c++){
   double output = outAct*cueWeight[c];
   Main.runs[subj].cues[c].retrieved += output;
   Input.totRetrCue.set(output+Input.totRetrCue.get());
  }
  for(int t=0; t<nTargets; t++){
   double output = outAct*targWeight[t];
   Main.runs[subj].targets[t].retrieved += output;
   Input.totRetrTarg.set(output+Input.totRetrTarg.get());
  }
  for(int d=0; d<nDistracters; d++){
   double output = outAct*distWeight[d];
   Main.runs[subj].distracters[d].retrieved += output;
   Input.totRetrDist.set(input+Input.totRetrDist.get());
  }
  for(int j=0; j<nGoals; j++){
   double output = outAct*goalWeight[j];
   Main.runs[subj].goals[j].retrieved += output;
   Goal.totRetrieved.set(output+Goal.totRetrieved.get());
  }
  for(int j=0; j<nDecisions; j++){
   double output = outAct*decWeight[j];
   Main.runs[subj].decisions[j].retrieved += output;
   Decision.totRetrieved.set(output+Decision.totRetrieved.get());
  }
  for(int j=0; j<nResponses; j++){
   double output = outAct*respWeight[j];
   Main.runs[subj].responses[j].retrieved += output;
   Response.totRetrieved.set(output+Response.totRetrieved.get());
  }
 }

 /**Global method for activating and connecting new nodes.
  * @param inflow Incoming activation.
  * @param weight Connection weight.
  * @return New weight.*/
 private double connect(double inflow, double weight){
  input += inflow * weight;
  if(inflow>weight){
   weight *= (1-writeConnect);
   weight += (inflow*writeConnect);
  }
  return weight;
 }

 /**Connects Episode to activated distracter Input nodes.
  * @param distIn Input from distracter Input node.
  * @param item Item number.*/
 public void connectDist(double distIn,int item){
  distWeight[item] = connect(distIn,distWeight[item]);
 }

 /**Connects Episode to activated target Input nodes.
  * @param targIn Input from target Input node.
  * @param item Item number.*/
 public void connectTarg(double targIn,int item){
  targWeight[item] = connect(targIn,targWeight[item]);
 }

 /**Connects Episode to activated cue Input nodes.
  * @param cueIn Input from cue Input node.
  * @param item Item number.*/
 public void connectCue(double cueIn,int item){
  cueWeight[item] = connect(cueIn,cueWeight[item]);
 }

 /**Connects Episode to activated Decision nodes.
  * @param decIn Input from Decision node.
  * @param item Item number.*/
 public void connectDec(double decIn,int item){
  decWeight[item] = connect(decIn,decWeight[item]);
 }

 /**Connects Episode to activated Goal nodes.
  * @param goalIn Input from Goal node.
  * @param item Item number.*/
 public void connectGoal(double goalIn,int item){
  goalWeight[item] = connect(goalIn,goalWeight[item]);
 }

 /**Connects Episode to activated Response nodes.
  * @param respIn Input from Response node.
  * @param item Item number.*/
 public void connectResp(double respIn,int item){
  respWeight[item] = connect(respIn,respWeight[item]);
 }

 /**Activates new Episode.*/
 public void activateNew(){
  input += .75; update();
 }

 /**Records response time.
  * @param respTime Response time.*/
 public void setRT(double respTime){
  rt = respTime;
 }

 /**Retrieves response time.
  * @param pace Current pace.
  * @return The amount to change the pace.*/
 public double getRT(double pace){
  double out = 0;
  if(activation>threshold){
   double diff = rt-pace;
   out = diff*timeWeight*(activation-threshold);
  }
  return out;
 }

 /**Adjusts weights in proportion to activation level.*/
 public void feedback(){
  double adjust = range(1-(peakAct*decayConnect));
  for(int item=0; item<nDistracters; item++) distWeight[item] *= adjust;
  for(int item=0; item<nTargets; item++) targWeight[item] *= adjust;
  for(int item=0; item<nCues; item++) cueWeight[item] *= adjust;
  for(int item=0; item<nGoals; item++) goalWeight[item] *= adjust;
  for(int item=0; item<nDecisions; item++) decWeight[item] *= adjust;
  for(int item=0; item<nResponses; item++) respWeight[item] *= adjust;
  timeWeight *= adjust;
  peakAct = 0;
 }

 /**Weakens episode weights after an error.*/
 public void tagError(){
  double adjust = .02;
  for(int item=0; item<nDistracters; item++) distWeight[item] *= adjust;
  for(int item=0; item<nTargets; item++) targWeight[item] *= adjust;
  for(int item=0; item<nCues; item++) cueWeight[item] *= adjust;
  for(int item=0; item<nGoals; item++) goalWeight[item] *= adjust;
  for(int item=0; item<nDecisions; item++) decWeight[item] *= adjust;
  for(int item=0; item<nResponses; item++) respWeight[item] *= adjust;
  timeWeight *= adjust;
 }

 /**Re-strengthens instruction memory.*/
 private void restore(){
  double adjust = .5;
  for(int item=0; item<nDecisions; item++)
   if(decWeight[item]>0) decWeight[item] += adjust;
  for(int item=0; item<nCues; item++)
   if(cueWeight[item]>0) cueWeight[item] += adjust;
  for(int item=0; item<nGoals; item++)
   if(goalWeight[item]>0) goalWeight[item] += adjust;
  for(int item=0; item<nResponses; item++)
   if(respWeight[item]>0) respWeight[item] += adjust;
 }

 /**Creates instructed memory (cue-goal).
  * @param id Identity node number.
  * @param dec Decision node number.*/
 public void setConnectGoal(int id,int goal){
  cueWeight[id] = 3;
  goalWeight[goal] = 3;
 }

 /**Creates instructed memory (decision-response).
  * @param dec Decision node number.
  * @param res Response node number.*/
 public void setConnectRes(int dec,int res){
  decWeight[dec] = 1;
  respWeight[res] = 1;
 }

 /**Creates instructed memory (stimulus-response).
  * @param stim Input node number.
  * @param res Response node number.*/
 public void setConnectSR(int stim,int res){
  targWeight[stim] = 1;
  decWeight[stim] = 1;
  respWeight[res] = 1;
 }

 /**Checks if cue was determined.
  * @param cue Cue number.*/
 public boolean wasCued(int cue){
  return (cueWeight[cue]>0) ? true : false;
 }

 /**Checks if decision was determined.
  * @param dec Decision number.*/
 public boolean wasDecided(int dec){
  return (decWeight[dec]>0) ? true : false;
 }

 /**Restores instructed cue-goal link.
  * @param item Item number.*/
 public void checkInstructCue(int item){
  if(cueWeight[item]>0) restore();
 }

 /**Restores instructed decision-response link.
  * @param item Item number.*/
 public void checkInstructDec(int item){
  if(decWeight[item]>0) restore();
 }

 /**Optional method to print episode weightings to output.*/
 public void printWeights(){
  DecimalFormat df = new DecimalFormat("#0.00");     
  System.out.println(df.format(4.0));
  for(int item=0; item<nDistracters; item++)
   System.out.print(""+df.format(distWeight[item])+"  ");
  System.out.println();
  for(int item=0; item<nTargets; item++)
   System.out.print(""+df.format(targWeight[item])+"  ");
  System.out.println();
  for(int item=0; item<nCues; item++)
   System.out.print(""+df.format(cueWeight[item])+"  ");
  System.out.println();
  for(int item=0; item<nGoals; item++)
   System.out.print(""+df.format(goalWeight[item])+"  ");
  System.out.println();
  for(int item=0; item<nDecisions; item++)
   System.out.print(""+df.format(decWeight[item])+"  ");
  System.out.println();
  for(int item=0; item<nResponses; item++)
   System.out.print(""+df.format(respWeight[item])+"  ");
  System.out.println(); System.out.println();
 }

 /**Optional method to return response weight.
  * @param index Response number.
  * @return Response weight.*/
 public double getWeight(int index){
  return respWeight[index];
 }
}